﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Windows.Forms;

namespace TestProject
{
    /// <summary>
    /// 메인 폼
    /// </summary>
    public partial class MainForm : Form
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 헬퍼
        /// </summary>
        private GeometryHepler helper = new GeometryHepler();

        /// <summary>
        /// 포인트 리스트
        /// </summary>
        private List<Point> pointList = new List<Point>();

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - MainForm()

        /// <summary>
        /// 생성자
        /// </summary>
        public MainForm()
        {
            InitializeComponent();

            #region 이벤트를 설정한다.

            MouseDown              += Form_MouseDown;
            Paint                  += Form_Paint;
            this.clearButton.Click += clearButton_Click;

            #endregion
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 폼 마우스 DOWN 처리하기 - Form_MouseDown(sender, e)

        /// <summary>
        /// 폼 마우스 DOWN 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void Form_MouseDown(object sender, MouseEventArgs e)
        {
            this.pointList.Add(new Point(e.X, e.Y));

            Invalidate();
        }

        #endregion
        #region 폼 페인트시 처리하기 - Form_Paint(sender, e)

        /// <summary>
        /// 폼 페인트시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void Form_Paint(object sender, PaintEventArgs e)
        {
            e.Graphics.Clear(BackColor);

            e.Graphics.SmoothingMode = SmoothingMode.AntiAlias;

            #region 마우스로 클릭한 포인트를 채워서 그린다.

            foreach(Point point in this.pointList)
            {
                e.Graphics.FillEllipse(Brushes.Cyan, point.X - 3, point.Y - 3, 7, 7);
            }

            #endregion

            List<Point> convexHullPointList = null;

            if(this.pointList.Count >= 3)
            {
                convexHullPointList = this.helper.GetConvexHullPointList(this.pointList);

                #region 컬링 사각형에 포함되지 않는 포인트를 그린다.

                foreach(Point point in this.helper.NonCulledPointArray)
                {
                    e.Graphics.FillEllipse(Brushes.White, point.X - 3, point.Y - 3, 7, 7);
                }

                #endregion
            }

            #region 마우스로 클릭한 포인트 외곽선을 그린다.

            foreach(Point point in this.pointList)
            {
                e.Graphics.DrawEllipse(Pens.Black, point.X - 3, point.Y - 3, 7, 7);
            }

            #endregion

            if(this.pointList.Count >= 3)
            {
                #region 최소/최대 포인트를 잇는 선을 그린다.

                e.Graphics.DrawPolygon(Pens.Red, this.helper.MinimumMaximumCornerPointArray);

                #endregion
                #region 컬링 사각형을 그린다.

                e.Graphics.DrawRectangle(Pens.Orange, this.helper.CullingRectangle);

                #endregion
                #region 외곽선을 그린다.

                Point[] convexHullPointArray = new Point[convexHullPointList.Count];

                convexHullPointList.CopyTo(convexHullPointArray);

                e.Graphics.DrawPolygon(Pens.Blue, convexHullPointArray);

                #endregion
            }
        }

        #endregion
        #region 지우기 버튼 클릭시 처리하기 - clearButton_Click(sender, e)

        /// <summary>
        /// 지우기 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void clearButton_Click(object sender, EventArgs e)
        {
            if(this.pointList != null)
            {
                this.pointList.Clear();
            }

            Invalidate();
        }

        #endregion
    }
}